/**
This script prints some working examples of the StrToBrokerTime() call into the experts-log.

17:36:11 mt4_timezone_support_test EURUSD,M5: loaded successfully
17:36:11 mt4_timezone_support_test EURUSD,M5 inputs: TZSupport.BrokerTimeZone_dsc="To configure the broker time zone. Mandatory"; TZSupport.BrokerTimeZone="Europe/London"; TZSupport.TimeZoneFileName_dsc1="This file is expected to reside in the experts/files/ folder"; TZSupport.TimeZoneFileName_dsc2="Source: ftp://ftp.iana.org/tz/releases/"; TZSupport.TimeZoneFileName="tzinfo.txt"; 
17:36:11 mt4_timezone_support_test EURUSD,M5: invalid integer number as parameter 1 for 'TimeToStr' function
17:36:11 mt4_timezone_support_test EURUSD,M5: Test 01:TimeCurrent() and "17:00 America/New York", "Europe/London") --> 1970.01.01 00:00:00   1373061600
17:36:11 mt4_timezone_support_test EURUSD,M5: Test 02:"2013-02-15 08:30" and "08:30 America/New York", "Europe/London") --> 2013.02.15 13:30:00
17:36:11 mt4_timezone_support_test EURUSD,M5: Test 03:"2013-03-15 08:30" and "08:30 America/New York", "Europe/London") --> 2013.03.15 12:30:00
17:36:11 mt4_timezone_support_test EURUSD,M5: Test 04:"2013-04-15 08:30" and "08:30 America/New York", "Europe/London")  --> 2013.04.15 13:30:00 Note the DST switch between America and Europe
17:36:11 mt4_timezone_support_test EURUSD,M5: Test 05:"2013-04-15 08:30" and "08:30 Australia/Sydney", "Europe/Berlin")  --> 2013.04.15 00:30:00
17:36:11 mt4_timezone_support_test EURUSD,M5: Test 06:"2013-03-15 08:30" and "08:30 Australia/Sydney", "Europe/London")  --> 2013.04.14 23:30:00 Note the date to roll back to 2013-04-14
17:36:11 mt4_timezone_support_test EURUSD,M5: Test 07:2013.04.15 23:31:00 Europe/London between "(16:30 America/New York" "08:45 Asia/Tokyo" ) gives 1 and 2013.04.15 21:30:00/2013.04.16 00:45:00
17:36:11 mt4_timezone_support_test EURUSD,M5: Test 08:2013.02.15 02:20:00 Europe/London between "(16:30 America/New York" "08:45 Asia/Tokyo" ) gives 0 and 2013.02.15 21:30:00/2013.02.15 23:45:00
17:36:11 mt4_timezone_support_test EURUSD,M5: Test 09:2013.02.15 08:20:00 EET between "(08:00 Europe/Berlin" "17:30 Europe/London" ) gives 0 and 2013.02.15 09:00:00/2013.02.15 19:30:00
17:36:11 mt4_timezone_support_test EURUSD,M5: uninit reason 0
17:36:11 mt4_timezone_support_test EURUSD,M5: removed
*/

#include <mt4_timezone_support.mqh>

int start() {
	// Initialize the time zone support object. 
	// This should normally be done in the init() procedure of the EA
	TZSupport.Init();
	
	// This demonstrates the convenience wrapper for the StrToBrokerTime() 
	// Note: The objecthandle and the reference time is omitted. The wrapper
	// replaces the objecthandle and the reference time by the global variable TZSupport.aHandle and
	// the second parameter by TimeCurrent()
	Print("Test 01:" 
		+ "TimeCurrent() and \"17:00 America/New York\", \"Europe/London\") --> "  
		+ TimeToStr( 
			TZSupport.StrToBrokerTime("17:00 America/New York", "Europe/London"), 
			TIME_DATE|TIME_MINUTES|TIME_SECONDS
		)
		+ "   " + TZSupport.StrToBrokerTime("17:00 America/New York", "Europe/London") 
	);
	
	// Here's a full parametrized direct call to the DLL function StrToBrokerTime().
	// Note that you have to pass an objecthandle and a reference time which is 
	// arbitrarily chooseable.
	// This example shows the difference between New York and London before the US DST switch.
	Print("Test 02:" 
		+ "\"2013-02-15 08:30\" and \"08:30 America/New York\", \"Europe/London\") --> "  
		+ TimeToStr( 
			StrToBrokerTime(TZSupport.aHandle, StrToTime("2013-02-15 12:30"), "08:30 America/New York", "Europe/London"),
			TIME_DATE|TIME_MINUTES|TIME_SECONDS
		)
	);
	
	// Here's a full parametrized direct call to the DLL function StrToBrokerTime().
	// Note that you have to pass an objecthandle and a reference time which is 
	// arbitrarily chooseable.
	// This example shows the difference between New York and London before the Europe DST switch.
	Print("Test 03:" 
		+ "\"2013-03-15 08:30\" and \"08:30 America/New York\", \"Europe/London\") --> "  
		+ TimeToStr( 
			StrToBrokerTime(TZSupport.aHandle, StrToTime("2013-03-15 12:30"), "08:30 America/New York", "Europe/London"),
			TIME_DATE|TIME_MINUTES|TIME_SECONDS
		)
	);
	
	// Here's a full parametrized direct call to the DLL function StrToBrokerTime().
	// Note that you have to pass an objecthandle and a reference time which is 
	// arbitrarily chooseable.
	// This example shows the difference between New York and London after the Europe DST switch.
	Print("Test 04:" 
		+ "\"2013-04-15 08:30\" and \"08:30 America/New York\", \"Europe/London\")  --> "   
		+ TimeToStr( 
			StrToBrokerTime(TZSupport.aHandle, StrToTime("2013-04-15 08:30"), "08:30 America/New York", "Europe/London"),
			TIME_DATE|TIME_MINUTES|TIME_SECONDS
		) 
		+ " Note the DST switch between America and Europe"
		
	);
	
	
	// Here's a full parametrized direct call to the DLL function StrToBrokerTime().
	// Note that you have to pass an objecthandle and a reference time which is 
	// arbitrarily chooseable.
	// This gives an example of the Australian trading session begin to a German based broker.
	Print("Test 05:" 
		+ "\"2013-04-15 08:30\" and \"08:30 Australia/Sydney\", \"Europe/Berlin\")  --> "  
		+ TimeToStr( 
			StrToBrokerTime(TZSupport.aHandle, StrToTime("2013-04-15 08:30"), "08:30 Australia/Sydney", "Europe/Berlin"),
			TIME_DATE|TIME_MINUTES|TIME_SECONDS
		)
		
	);
	
	// Here's a full parametrized direct call to the DLL function StrToBrokerTime().
	// Note that you have to pass an objecthandle and a reference time which is 
	// arbitrarily chooseable.
	// This gives an example of the Australian trading session begin to a UK based broker.
	Print("Test 06:" 
		+ "\"2013-03-15 08:30\" and \"08:30 Australia/Sydney\", \"Europe/London\")  --> "  
		+ TimeToStr( 
			StrToBrokerTime(TZSupport.aHandle, StrToTime("2013-04-15 08:30"), "08:30 Australia/Sydney", "Europe/London"),
			TIME_DATE|TIME_MINUTES|TIME_SECONDS
		)
		+ " Note the date to roll back to 2013-04-14"
	);
	
	// Here's a call to check whether a given time x is between two time boundaries where (pFromTime <= x < pUntilTime)
	// This is usable to filter trades by a given time window which e.g. is defined 
	// as "16:30 New/York" (read: Trading ends in NY) and "08:45 Asia/Tokyo" (read: "Trading starts in Tokyo")
	// The results of the recalculations of the timeStr are stored in a datetime[2] array.
	// You MUST pass a datetime[2] array to the DLL call as MQL4 does not allow single 
	// datetime values passed to DLL functions by reference.
	string 	pTimeFromStr="16:30 America/New York", 
			pTimeUntilStr="08:45 Asia/Tokyo",
			pBrokerTimeZone="Europe/London";
	datetime wrapperDateTime[2];
	datetime pBrokerTime=StrToTime("2013-04-15 23:31");
	int rc=isTimeBetweenRangeDLL(
			TZSupport.aHandle,
			pBrokerTime,
			wrapperDateTime,
			pTimeFromStr, pTimeUntilStr,
			pBrokerTimeZone
	);
	Print("Test 07:" 
		+ TimeToStr( pBrokerTime, TIME_DATE|TIME_MINUTES|TIME_SECONDS) + " " + pBrokerTimeZone + " between "
		+ "\"(" + pTimeFromStr + "\" "
		+ "\"" + pTimeUntilStr + "\" )"
		+ " gives " + rc + " and " 
		+ TimeToStr( wrapperDateTime[0], TIME_DATE|TIME_MINUTES|TIME_SECONDS) + "/"
		+ TimeToStr( wrapperDateTime[1], TIME_DATE|TIME_MINUTES|TIME_SECONDS)
	);
	
	// This gives an example of a *really* weird DST switch.
	pBrokerTime=StrToTime("2013-02-15 02:20");
	rc=isTimeBetweenRangeDLL(
			TZSupport.aHandle,
			pBrokerTime,
			wrapperDateTime,
			pTimeFromStr, pTimeUntilStr,
			pBrokerTimeZone
	);
	Print("Test 08:" 
		+ TimeToStr( pBrokerTime, TIME_DATE|TIME_MINUTES|TIME_SECONDS) + " " + pBrokerTimeZone + " between "
		+ "\"(" + pTimeFromStr + "\" "
		+ "\"" + pTimeUntilStr + "\" )"
		+ " gives " + rc + " and " 
		+ TimeToStr( wrapperDateTime[0], TIME_DATE|TIME_MINUTES|TIME_SECONDS) + "/"
		+ TimeToStr( wrapperDateTime[1], TIME_DATE|TIME_MINUTES|TIME_SECONDS)
	);
	
	// This is an example to translate the Frankfurt Open and the London Close into the Alpari time zone (EET)
	pBrokerTimeZone="EET";
	pTimeFromStr="08:00 Europe/Berlin";
	pTimeUntilStr="17:30 Europe/London";
	pBrokerTime=StrToTime("2013-02-15 08:20");
	rc=isTimeBetweenRangeDLL(
			TZSupport.aHandle,
			pBrokerTime,
			wrapperDateTime,
			pTimeFromStr, pTimeUntilStr,
			pBrokerTimeZone
	);
	Print("Test 09:" 
		+ TimeToStr( pBrokerTime, TIME_DATE|TIME_MINUTES|TIME_SECONDS) + " " + pBrokerTimeZone + " between "
		+ "\"(" + pTimeFromStr + "\" "
		+ "\"" + pTimeUntilStr + "\" )"
		+ " gives " + rc + " and " 
		+ TimeToStr( wrapperDateTime[0], TIME_DATE|TIME_MINUTES|TIME_SECONDS) + "/"
		+ TimeToStr( wrapperDateTime[1], TIME_DATE|TIME_MINUTES|TIME_SECONDS)
	);
	
	// Free the memory used by the time zone support object.
	// This should normally be done in the deinit() procedure of the EA.
	TZSupport.DeInit();
	return(0);
}

